"""
This module manages language support for Amulet.
The language files are in amulet_map_editor/lang with RFC 1766 file names and the .lang extension.
Lines starting with a hash (#) are ignored as comments and blank lines are ignored.
All other lines should have the format `key=value` where `key` is the language key used in the application and `value` is the localised string.
Keys should match the regex [a-z.]+ (lower case a-z and the full stop character) and values can contain any unicode character.
Files must be utf-8 encoded.
"""
import glob
from typing import Dict, Optional, List, Tuple
import os
import locale
import logging
import re

import amulet_map_editor
from amulet_map_editor.api import config as CONFIG

log = logging.getLogger(__name__)

# there might be a proper way to do this but this should be enough for now

_lang_dirs: List[str] = []  # the language directories
_lang: Dict[
    str, str
] = {}  # a storage for the language strings. unique_identifier: language_string

_default_language = "en"
_active_language: Optional[str] = None


def lang_dirs() -> Tuple[str, ...]:
    """Tuple of known language directories."""
    return tuple(_lang_dirs)


def parse_language_id(language_id: str) -> Optional[Tuple[str, str]]:
    """
    Parse an RFC 1766 language id string into a nicer format.

    :param language_id: The RFC 1766 code to parse
    :return: A tuple of the language and region codes (both lower case). The region code may be None.
    """
    match = re.fullmatch(
        r"(?P<language>[a-z]+)([-_](?P<region>[a-z]+))?.*",
        language_id,
        flags=re.IGNORECASE,
    )
    if match is None:
        return None
    region = match.group("region")
    region = "" if region is None else region.lower()
    return match.group("language").lower(), region


def set_language(language_id: str):
    """
    Sets and loads the specified language.
    A restart may be required.

    :param language_id: The RFC 1766 language code to use
    """
    global _active_language
    _active_language = language_id
    cfg = CONFIG.get("amulet_meta", {})
    cfg["lang"] = language_id
    CONFIG.put("amulet_meta", cfg)
    _load_language()


def get_language() -> str:
    """Get the currently active language string."""
    global _active_language
    if _active_language is None:
        # find out what language the user is using.
        try:
            # try getting the OS language
            _active_language = locale.getdefaultlocale()[0]
        except:
            # if that fails use the default language
            _active_language = _default_language

        # if a language is set in the config use that
        _active_language = CONFIG.get("amulet_meta", {}).get("lang", _active_language)

        if _active_language is None:
            _active_language = _default_language
    return _active_language


def get_languages() -> List[str]:
    """Get a list of all supported language codes."""
    langs = set()
    for d in _lang_dirs:
        for l in glob.glob(os.path.join(glob.escape(d), "*.lang")):
            langs.add(os.path.basename(l)[:-5])
    return sorted(langs)


def _load_lang_file(lang_path: str) -> Dict[str, str]:
    """Loads a language file and returns the result as a dictionary.
    Will return an empty dictionary if the path does not exist.

    :param lang_path: The lang file path to load.
    :return: A dictionary mapping the string identifier to the string
    """
    lang = {}
    if not os.path.isfile(lang_path):
        return lang
    with open(lang_path, encoding="utf-8") as f:
        for line in f.readlines():
            lstrip_line = line.lstrip()
            if lstrip_line.startswith("#") or not lstrip_line:
                continue
            split_line = lstrip_line.split("=", 1)
            if len(split_line) == 2:
                unique_identifier = split_line[0].rstrip()
                language_string = split_line[1].replace("\\n", "\n").strip()
                lang[unique_identifier] = language_string
    return lang


def _find_langs(path: str) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """Find the default, language specific and region specific lang paths."""
    langs = {
        parse_language_id(os.path.basename(lpath)[:-5]): lpath
        for lpath in glob.glob(os.path.join(glob.escape(path), "*.lang"))
    }

    default_key = parse_language_id(_default_language)
    active_key = parse_language_id(get_language())

    default_path = langs.get(default_key, None)
    language_path = None
    region_path = None

    if active_key is not None:
        if active_key[0] != default_key[0]:
            language_path = langs.get((active_key[0], ""), None)
        if active_key[1]:
            region_path = langs.get(active_key, None)

    return default_path, language_path, region_path


def _load_language():
    """Load the translations for the active language."""
    # work out which lang files to load based on the active language
    # lang files from the default langauge
    lang_files = sum(zip(*list(map(_find_langs, lang_dirs()))), ())

    _lang.clear()
    for lang_file in lang_files:
        if lang_file is None:
            continue
        _lang.update(_load_lang_file(lang_file))


def register_lang_directory(lang_dir: str):
    """Register a new language directory.
    Use this, if you are a third party program, to register new translations for your program.

    :param lang_dir: The directory containing language files to register. Files must be of the format <language_code>.lang eg en_US.lang
    :return:
    """
    if os.path.isdir(lang_dir):
        if lang_dir in _lang_dirs:
            log.warning(
                f"The language directory {lang_dir} has already been registered."
            )
        else:
            _lang_dirs.append(lang_dir)
            _load_language()


# load the normal language directory
register_lang_directory(
    os.path.join(os.path.dirname(amulet_map_editor.__file__), "lang")
)


def get(unique_identifier: str):
    """
    Get the localised string for the translation key

    :param unique_identifier: The translation key to get.
    :return: The translated string.
    """
    if unique_identifier not in _lang:
        # help debugging referenced lang entries that do not exist
        log.info(f"Could not find lang entry for {unique_identifier}")
    return _lang.get(unique_identifier, unique_identifier)



import os

from amulet_map_editor.programs.edit.plugins.operations import stock_plugins


STOCK_PLUGINS_DIR: str = stock_plugins.__path__[0]
STOCK_PLUGINS_NAME: str = stock_plugins.__name__
CUSTOM_PLUGINS_DIR = os.path.join(os.environ["DATA_DIR"], "plugins")



import os
import glob
import string
import unittest
from amulet_map_editor.api.lang import lang_dirs


KeyCharacterSet = set(string.ascii_lowercase + string.digits + "_.")


class LangTestCase(unittest.TestCase):
    def test_lang(self):
        for lang_dir in lang_dirs():
            for lang_path in glob.glob(os.path.join(glob.escape(lang_dir), "*")):
                self.assertTrue(
                    lang_path.endswith(".lang"),
                    f'{lang_path} does not end with ".lang".',
                )
                self.assertTrue(
                    os.path.isfile(lang_path), f"{lang_path} is not a file."
                )
                with open(lang_path, encoding="utf-8") as f:
                    unique_ids = set()
                    for line_number, line in enumerate(f.readlines()):
                        line_number += 1
                        strip_line = line.strip()
                        if strip_line and not strip_line.startswith("#"):
                            split_line = line.split("=", 1)
                            self.assertEqual(
                                len(split_line),
                                2,
                                f"{lang_path} line {line_number}: No = symbol present.",
                            )
                            unique_identifier = split_line[0].strip()
                            self.assertTrue(
                                len(unique_identifier) > 0,
                                f"{lang_path} line {line_number}: There is no identifier before the first = symbol.",
                            )
                            self.assertTrue(
                                set(unique_identifier).issubset(KeyCharacterSet),
                                f"{lang_path} line {line_number}: identifier is not a subset of a-z0-9_.",
                            )
                            self.assertTrue(
                                unique_identifier not in unique_ids,
                                f"{lang_path} line {line_number}: identifier already defined in the file.",
                            )
                            unique_ids.add(unique_identifier)
                            language_string = split_line[1].replace("\\n", "\n").strip()
                            self.assertTrue(
                                len(language_string) > 0,
                                f"{lang_path} line {line_number}: There is no translation string after the first = symbol.",
                            )


if __name__ == "__main__":
    unittest.main()



from typing import TYPE_CHECKING, Optional
import wx
from OpenGL.GL import (
    glClear,
    GL_COLOR_BUFFER_BIT,
    GL_DEPTH_BUFFER_BIT,
)

from amulet.api.data_types import Dimension

from amulet_map_editor.api.opengl.camera import Projection
from amulet_map_editor.api.opengl.mesh.level import RenderLevel
from amulet_map_editor.api.opengl.mesh.level_group import LevelGroup
from amulet_map_editor.api.opengl.mesh.sky_box import SkyBox
from amulet_map_editor.api.opengl.resource_pack.resource_pack import OpenGLResourcePack

from .chunk_generator import ThreadingEnabled, ChunkGenerator
from .edit_canvas_container import EditCanvasContainer
from .events import (
    DimensionChangeEvent,
    CameraMovedEvent,
    EVT_CAMERA_MOVED,
    PreDrawEvent,
    DrawEvent,
    PostDrawEvent,
)

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class Renderer(EditCanvasContainer):
    """This class holds the drawable objects and has methods to draw them."""

    __slots__ = (
        "_render_distance",
        "_chunk_generator",
        "_opengl_resource_pack",
        "_render_world",
        "_chunk_generator",
        "_fake_levels",
        "_sky_box",
        "_draw_timer",
        "_gc_timer",
    )

    _sky_box: Optional[SkyBox]
    _fake_levels: Optional[LevelGroup]

    def __init__(
        self,
        canvas: "EditCanvas",
        world,
        context_identifier: str,
        opengl_resource_pack: OpenGLResourcePack,
    ):
        super().__init__(canvas)
        self._render_distance = 5

        self._chunk_generator = ChunkGenerator()
        self._opengl_resource_pack = opengl_resource_pack

        self._render_world = RenderLevel(
            context_identifier,
            opengl_resource_pack,
            world,
            draw_floor=True,
            draw_ceil=True,
        )
        self._chunk_generator.register(self._render_world)

        self._fake_levels = None
        self._sky_box = None

        self._draw_timer = wx.Timer(self.canvas)
        self._gc_timer = wx.Timer(self.canvas)

    def bind_events(self):
        """Set up all events required to run."""
        self.canvas.Bind(wx.EVT_TIMER, self._gc, self._gc_timer)
        self.canvas.Bind(
            wx.EVT_TIMER,
            self._do_draw,
            self._draw_timer,
        )
        self.canvas.Bind(EVT_CAMERA_MOVED, self._on_camera_moved)
        self.canvas.Bind(wx.EVT_WINDOW_DESTROY, self._on_destroy, self.canvas)

    def enable(self):
        """Enable and start working."""
        self.enable_threads()

    def disable(self):
        """Disable and unload all geometry."""
        self.disable_threads()
        self.render_world.unload()
        self.fake_levels.unload()

    def _on_destroy(self, evt):
        self.disable()
        evt.Skip()

    def is_closeable(self):
        """Check that the data is safe to be closed."""
        return self.render_world.is_closeable()

    def close(self):
        """Close and destroy all data."""
        self.render_world.close()
        self.fake_levels.clear()
        self.sky_box.unload()

    def disable_threads(self):
        """Stop the generation of new chunk geometry.
        Makes it safe to modify the world data."""
        self._draw_timer.Stop()
        self._gc_timer.Stop()
        self._chunk_generator.stop()

    def enable_threads(self):
        """Start the generation of new chunk geometry."""
        self.render_world.enable()
        self.fake_levels.enable()
        self._chunk_generator.start()
        self._draw_timer.Start(15)
        self._gc_timer.Start(10000)

    # TODO: move this logic into a resource pack reload method
    # def _load_resource_pack(self, *resource_packs: JavaResourcePack):
    #     self._resource_pack = JavaResourcePackManager(resource_packs)
    #     for _ in self._create_atlas():
    #         pass

    @property
    def opengl_resource_pack(self) -> OpenGLResourcePack:
        return self._opengl_resource_pack

    @property
    def render_world(self) -> RenderLevel:
        return self._render_world

    @property
    def fake_levels(self) -> LevelGroup:
        """Floating levels that are not the main level."""
        if self._fake_levels is None:
            self._fake_levels: LevelGroup = LevelGroup(
                self.canvas.context_identifier,
                self.opengl_resource_pack,
            )
            self._chunk_generator.register(self._fake_levels)
        return self._fake_levels

    @property
    def sky_box(self) -> SkyBox:
        """The cube in the distance displaying the sky."""
        if self._sky_box is None:
            self._sky_box = SkyBox(
                self._render_world.context_identifier,
                self._opengl_resource_pack,
            )
        return self._sky_box

    @property
    def dimension(self) -> Dimension:
        """The currently loaded dimension in the renderer."""
        return self.render_world.dimension

    @dimension.setter
    def dimension(self, dimension: Dimension):
        """Set the currently loaded dimension in the renderer."""
        if dimension != self.dimension:
            self.disable_threads()
            self.render_world.dimension = dimension
            wx.PostEvent(self.canvas, DimensionChangeEvent(dimension=dimension))
            self.enable_threads()

    @property
    def render_distance(self) -> int:
        """The distance from the camera in chunks that should be drawn"""
        return self._render_distance

    @render_distance.setter
    def render_distance(self, render_distance: int):
        """Set the distance from the camera in chunks that should be drawn"""
        self._render_distance = render_distance
        self.render_world.render_distance = render_distance
        # self.fake_levels.render_distance = render_distance  # TODO

    def _on_camera_moved(self, evt: CameraMovedEvent):
        """The camera has moved. Update each class's camera state."""
        self.move_camera(evt.camera_location, evt.camera_rotation)
        evt.Skip()

    def move_camera(self, location, rotation):
        # TODO: add combined methods
        self.render_world.camera_location = location
        self.render_world.camera_rotation = rotation

        self.fake_levels.set_camera_location(*location)
        self.fake_levels.set_camera_rotation(*rotation)

        self.sky_box.set_camera_location(location)

    def _do_draw(self, evt):
        wx.PostEvent(self.canvas, PreDrawEvent())
        wx.PostEvent(self.canvas, DrawEvent())
        wx.PostEvent(self.canvas, PostDrawEvent())

    def default_draw(self):
        """The default draw logic."""
        self.start_draw()
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self.draw_sky_box()
            glClear(GL_DEPTH_BUFFER_BIT)
        self.draw_level()
        self.end_draw()

    def start_draw(self):
        """Run commands before drawing."""
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

    def draw_sky_box(self):
        """Draw the skybox."""
        self.sky_box.draw(self.canvas.camera.transformation_matrix)

    def draw_level(self):
        """Draw the main level."""
        self.render_world.draw(self.canvas.camera.transformation_matrix)

    def draw_fake_levels(self):
        """Draw the floating structure levels."""
        self.fake_levels.draw(self.canvas.camera.transformation_matrix)

    if ThreadingEnabled:

        def end_draw(self):
            """Run commands after drawing."""
            self.canvas.SwapBuffers()

    else:

        def end_draw(self):
            """Run commands after drawing."""
            self.canvas.SwapBuffers()
            self._chunk_generator.thread_action()

    def _gc(self, event):
        """Unload data to limit memory usage."""
        self.render_world.run_garbage_collector()
        self.fake_levels.run_garbage_collector()
        event.Skip()



import os
import platformdirs

experimental_bedrock_resources = False

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

# Initialise default paths. These should be initialised in main before this is imported
# but the tests don't run main so a default is required for that case.
data_dir = platformdirs.user_data_dir("AmuletMapEditor", "AmuletTeam")
os.environ.setdefault("DATA_DIR", data_dir)
config_dir = platformdirs.user_config_dir("AmuletMapEditor", "AmuletTeam")
if config_dir == data_dir:
    config_dir = os.path.join(data_dir, "Config")
os.environ.setdefault("CONFIG_DIR", config_dir)
os.environ.setdefault(
    "CACHE_DIR", platformdirs.user_cache_dir("AmuletMapEditor", "AmuletTeam")
)
os.environ.setdefault(
    "LOG_DIR", platformdirs.user_log_dir("AmuletMapEditor", "AmuletTeam")
)

from amulet_map_editor.api import config as CONFIG, lang
from amulet_map_editor.api.framework.app import open_level, close_level


# This is here because of the major change in the NBT library
# We have tried updating all the code that will break, but we will inevitably miss some
# This modifies the behaviour of warnings.warn to notify us if the old functionality is still used so we can fix it.
def _patch_warn():
    from threading import Thread
    import warnings
    import urllib.request
    import urllib.parse
    from urllib.error import URLError
    import traceback
    import os
    import getpass

    stacks = set()

    def _call_home(message, category, stack):
        try:
            stack = stack.replace(
                os.getcwd(), "."
            )  # remove the parent path. We don't need it
            stack = stack.replace(
                getpass.getuser(), "%username%"
            )  # Remove the username just in case
        except:
            pass

        if stack not in stacks:
            stacks.add(stack)
            try:
                urllib.request.urlopen(
                    f"https://docs.google.com/forms/d/e/1FAIpQLSfYb44EYR78yJlQAgyf-GTWn1ZlaGNMuIerl_XTpaYN0d8B4Q/formResponse?"
                    f"entry.601089147={urllib.parse.quote(__version__)}&"
                    f"entry.961557167={urllib.parse.quote(message)}&"
                    f"entry.2137739277={urllib.parse.quote(str(category))}&"
                    f"entry.1185835095={urllib.parse.quote(stack)}",
                    timeout=5,
                )
            except URLError:
                pass

    _old_warn = warnings.warn

    def _on_warn(message, category=None, **kwargs):
        _old_warn(message, category, **kwargs)
        t = Thread(
            target=_call_home,
            args=(message, category, "".join(traceback.format_stack(limit=6))),
        )
        t.start()

    warnings.warn = _on_warn


_patch_warn()



import wx
from typing import TYPE_CHECKING, Optional
import traceback
import logging

import os
import sys
import subprocess

from amulet_map_editor.api import image
from amulet_map_editor.api.wx.ui.simple import SimpleChoiceAny
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog

from amulet_map_editor.programs.edit.api.operations import OperationUIType
from amulet_map_editor.programs.edit.api.operations.manager import UIOperationManager
from .base_tool_ui import BaseToolUI

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas

log = logging.getLogger(__name__)


class BaseOperationChoiceToolUI(wx.BoxSizer, BaseToolUI):
    OperationGroupName = None
    _active_operation: Optional[OperationUIType]

    ShowOpenFolder = True

    def __init__(self, canvas: "EditCanvas"):
        wx.BoxSizer.__init__(self, wx.VERTICAL)
        BaseToolUI.__init__(self, canvas)

        self._active_operation: Optional[OperationUIType] = None
        self._last_active_operation_id: Optional[str] = None

        horizontal_sizer = wx.BoxSizer(wx.HORIZONTAL)

        assert isinstance(
            self.OperationGroupName, str
        ), "OperationGroupName has not been set or is not a string."
        # The operation selection
        self._operation_choice = SimpleChoiceAny(self.canvas)
        horizontal_sizer.Add(self._operation_choice)
        self._operations = UIOperationManager(self.OperationGroupName)
        self._operation_choice.SetItems(
            {op.identifier: op.name for op in self._operations.operations}
        )
        self._operation_choice.Bind(wx.EVT_CHOICE, self._on_operation_change)

        # The reload button
        self._reload_operation = wx.BitmapButton(
            self.canvas, bitmap=image.REFRESH_ICON.bitmap(16, 16)
        )
        self._reload_operation.SetToolTip("Reload Operations")
        horizontal_sizer.Add(self._reload_operation)
        self._reload_operation.Bind(wx.EVT_BUTTON, self._on_reload_operations)

        # The open folder button
        if self.ShowOpenFolder:
            self._open_folder = wx.BitmapButton(
                self.canvas, bitmap=image.TABLERICONS.folder.bitmap(16, 16)
            )
            self._open_folder.SetToolTip("Open Plugin Folder")
            horizontal_sizer.Add(self._open_folder)
            self._open_folder.Bind(wx.EVT_BUTTON, self._on_open_folder)

        self.Add(horizontal_sizer)

        self._operation_sizer = wx.BoxSizer(wx.VERTICAL)
        self.Add(self._operation_sizer, 1, wx.EXPAND)

    @property
    def name(self) -> str:
        """The name of the group of operations."""
        raise NotImplementedError

    @property
    def active_operation_id(self) -> str:
        """The identifier of the operation selected by the choice input.
        Note if in the process of changing this may be different to self._active_operation.
        """
        return self._operation_choice.GetCurrentObject()

    def _on_operation_change(self, evt):
        """Run when the operation selection changes."""
        if (
            self.active_operation_id
            and self._last_active_operation_id != self.active_operation_id
        ):
            self._setup_operation()
            self.canvas.reset_bound_events()
        evt.Skip()

    def _setup_operation(self):
        """Remove the old operation and create the UI for the new operation."""
        operation_path = self.active_operation_id
        if operation_path:
            # only reload the operation if the
            operation = self._operations[operation_path]
            if self._active_operation is not None:
                self._active_operation.disable()
            self._operation_sizer.Clear(delete_windows=True)
            try:
                self._active_operation = operation(
                    self.canvas, self.canvas, self.canvas.world
                )
                self._operation_sizer.Add(
                    self._active_operation, *self._active_operation.wx_add_options
                )
                self._active_operation.enable()
            except Exception as e:
                # If something went wrong clear the created UI
                self._active_operation = None
                self._operation_sizer.Clear(delete_windows=True)
                for window in self.canvas.GetChildren():
                    window: wx.Window
                    # remove orphaned windows.
                    # If the Sizer.Add method was not run it will not be in self._operation_sizer
                    if window.GetContainingSizer() is None:
                        window.Destroy()
                log.error("Error loading Operation UI.", exc_info=True)
                dialog = TracebackDialog(
                    self.canvas,
                    "Error loading Operation UI.",
                    str(e),
                    traceback.format_exc(),
                )
                dialog.ShowModal()
                dialog.Destroy()
            finally:
                self._last_active_operation_id = operation.identifier
                self.Layout()

    def bind_events(self):
        if self._active_operation is not None:
            self._active_operation.bind_events()

    def enable(self):
        if self._active_operation is None:
            self._setup_operation()
        else:
            self._active_operation.enable()

    def disable(self):
        if self._active_operation is not None:
            self._active_operation.disable()

    def _on_reload_operations(self, evt):
        """Run when the button is pressed to reload the operations."""
        self.reload_operations()

    def reload_operations(self):
        """Reload all operations and repopulate the UI."""
        # store the id of the old operation
        operation_id = self.active_operation_id

        # reload the operations
        self._operations.reload()

        # repopulate the selection
        self._operation_choice.SetItems(
            {op.identifier: op.name for op in self._operations.operations}
        )

        if operation_id:
            identifiers = self._operation_choice.values

            if identifiers:
                if operation_id in identifiers:
                    self._operation_choice.SetSelection(identifiers.index(operation_id))
                else:
                    log.info(f"Operation {operation_id} was not found.")
                    self._operation_choice.SetSelection(0)
            else:
                log.error("No operations found. Something has gone wrong.")

            self._setup_operation()
            self.canvas.reset_bound_events()

    def _on_open_folder(self, evt):
        path = self._operations.public_path
        if not os.path.exists(path):
            os.makedirs(path)
        if sys.platform == "win32":
            os.startfile(path)
        else:
            opener = "open" if sys.platform == "darwin" else "xdg-open"
            subprocess.call([opener, path])



# API layout
from .base_behaviour import BaseBehaviour

# camera behaviour
from .camera_behaviour import CameraBehaviour

# ray casting behaviour
from .raycast_behaviour import RaycastBehaviour

# pointer box behaviour
from .pointer_behaviour import PointerBehaviour

# static selection
from .static_selection_behaviour import StaticSelectionBehaviour

# block selection
from .block_selection_behaviour import BlockSelectionBehaviour

# chunk selection
from .chunk_selection_behaviour import ChunkSelectionBehaviour



from typing import TYPE_CHECKING, Optional
import wx

from amulet_map_editor.programs.edit.api.edit_canvas_container import (
    EditCanvasContainer,
)
from amulet_map_editor.api.wx.ui.simple import SimpleChoiceAny
from amulet_map_editor.programs.edit.api.events import (
    EVT_CAMERA_MOVED,
    EVT_SPEED_CHANGED,
    EVT_UNDO,
    EVT_REDO,
    EVT_CREATE_UNDO,
    EVT_SAVE,
    EVT_PROJECTION_CHANGED,
    EVT_DIMENSION_CHANGE,
    DimensionChangeEvent,
    EditCloseEvent,
)
from amulet_map_editor.api import image, lang
from amulet_map_editor.api.opengl.camera import Projection

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class FilePanel(wx.BoxSizer, EditCanvasContainer):
    def __init__(self, canvas: "EditCanvas"):
        wx.BoxSizer.__init__(self, wx.HORIZONTAL)
        EditCanvasContainer.__init__(self, canvas)

        level = self.canvas.world
        self._version_text = wx.StaticText(
            canvas,
            label=f"{level.level_wrapper.platform}, {level.level_wrapper.version}",
        )
        self._version_text.SetToolTip(
            lang.get("program_3d_edit.file_ui.version_tooltip")
        )
        self.Add(self._version_text, 0)
        self.AddStretchSpacer(1)
        self._projection_button = wx.Button(canvas, label="3D")
        self._projection_button.SetToolTip(
            lang.get("program_3d_edit.file_ui.projection_tooltip")
        )
        self._projection_button.Bind(wx.EVT_BUTTON, self._on_projection_button)
        self.Add(
            self._projection_button, 0, wx.TOP | wx.BOTTOM | wx.RIGHT | wx.CENTER, 5
        )
        self._location_button = wx.Button(
            canvas, label=", ".join([f"{s:.2f}" for s in self.canvas.camera.location])
        )
        self._location_button.SetToolTip(
            lang.get("program_3d_edit.file_ui.location_tooltip")
        )
        self._location_button.Bind(wx.EVT_BUTTON, lambda evt: self.canvas.goto())
        self.Add(self._location_button, 0, wx.TOP | wx.BOTTOM | wx.RIGHT | wx.CENTER, 5)

        def set_speed(evt):
            dialog = SpeedSelectDialog(
                canvas, self.canvas.camera.move_speed * 1000 / 33
            )
            if dialog.ShowModal() == wx.ID_OK:
                self.canvas.camera.move_speed = dialog.speed * 33 / 1000

        self._speed_button = wx.Button(
            canvas,
            label=f"{self.canvas.camera.move_speed*1000/33:.4g}{lang.get('program_3d_edit.file_ui.speed_blocks_per_second')}",
        )
        self._speed_button.SetToolTip(lang.get("program_3d_edit.file_ui.speed_tooltip"))
        self._speed_button.Bind(wx.EVT_BUTTON, set_speed)
        self.Add(self._speed_button, 0, wx.TOP | wx.BOTTOM | wx.RIGHT | wx.CENTER, 5)

        self._dim_options = SimpleChoiceAny(canvas)
        self._dim_options.SetToolTip(lang.get("program_3d_edit.file_ui.dim_tooltip"))
        self._dim_options.SetItems(level.level_wrapper.dimensions)
        self._dim_options.SetValue(level.level_wrapper.dimensions[0])
        self._dim_options.Bind(wx.EVT_CHOICE, self._on_dimension_change)

        self.Add(self._dim_options, 0, wx.TOP | wx.BOTTOM | wx.RIGHT | wx.CENTER, 5)

        def create_button(text, operation):
            button = wx.Button(canvas, label=text)
            button.Bind(wx.EVT_BUTTON, operation)
            self.Add(button, 0, wx.TOP | wx.BOTTOM | wx.RIGHT, 5)
            return button

        self._undo_button: Optional[wx.Button] = create_button(
            "0", lambda evt: self.canvas.undo()
        )
        self._undo_button.SetBitmap(image.icon.tablericons.arrow_back_up.bitmap(20, 20))
        self._undo_button.SetToolTip(lang.get("program_3d_edit.file_ui.undo_tooltip"))

        self._redo_button: Optional[wx.Button] = create_button(
            "0", lambda evt: self.canvas.redo()
        )
        self._redo_button.SetBitmap(
            image.icon.tablericons.arrow_forward_up.bitmap(20, 20)
        )
        self._redo_button.SetToolTip(lang.get("program_3d_edit.file_ui.redo_tooltip"))

        self._save_button: Optional[wx.Button] = create_button(
            "0", lambda evt: self.canvas.save()
        )
        self._save_button.SetBitmap(image.icon.tablericons.device_floppy.bitmap(20, 20))
        self._save_button.SetToolTip(lang.get("program_3d_edit.file_ui.save_tooltip"))

        close_button = wx.BitmapButton(
            canvas, bitmap=image.icon.tablericons.square_x.bitmap(20, 20)
        )
        close_button.SetToolTip(lang.get("program_3d_edit.file_ui.close_tooltip"))
        close_button.Bind(
            wx.EVT_BUTTON, lambda evt: wx.PostEvent(self.canvas, EditCloseEvent())
        )
        self.Add(close_button, 0, wx.TOP | wx.BOTTOM | wx.RIGHT, 5)

        self._update_buttons()

        self.Layout()

    def bind_events(self):
        self.canvas.Bind(EVT_CAMERA_MOVED, self._on_camera_move)
        self.canvas.Bind(EVT_SPEED_CHANGED, self._on_speed_change)
        self.canvas.Bind(EVT_UNDO, self._on_update_buttons)
        self.canvas.Bind(EVT_REDO, self._on_update_buttons)
        self.canvas.Bind(EVT_SAVE, self._on_update_buttons)
        self.canvas.Bind(EVT_CREATE_UNDO, self._on_update_buttons)
        self.canvas.Bind(EVT_PROJECTION_CHANGED, self._on_projection_change)
        self.canvas.Bind(EVT_DIMENSION_CHANGE, self._change_dimension)

    def _on_update_buttons(self, evt):
        self._update_buttons()
        evt.Skip()

    def _update_buttons(self):
        self._undo_button.SetLabel(f"{self.canvas.world.history_manager.undo_count}")
        self._redo_button.SetLabel(f"{self.canvas.world.history_manager.redo_count}")
        self._save_button.SetLabel(
            f"{self.canvas.world.history_manager.unsaved_changes}"
        )

    def _on_dimension_change(self, evt):
        """Run when the dimension selection is changed by the user."""
        dimension = self._dim_options.GetCurrentObject()
        if dimension is not None:
            self.canvas.dimension = dimension
        evt.Skip()

    def _on_projection_change(self, evt):
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self._projection_button.SetLabel("3D")
        elif self.canvas.camera.projection_mode == Projection.TOP_DOWN:
            self._projection_button.SetLabel("2D")
        evt.Skip()

    def _on_projection_button(self, evt):
        if self.canvas.camera.projection_mode == Projection.PERSPECTIVE:
            self.canvas.camera.projection_mode = Projection.TOP_DOWN
        else:
            self.canvas.camera.projection_mode = Projection.PERSPECTIVE
        evt.Skip()

    def _change_dimension(self, evt: DimensionChangeEvent):
        """Run when the dimension attribute in the canvas is changed.
        This is run when the user changes the attribute and when it is changed manually in code.
        """
        dimension = evt.dimension
        index = self._dim_options.FindString(dimension)
        if not (index == wx.NOT_FOUND or index == self._dim_options.GetSelection()):
            self._dim_options.SetSelection(index)

    def _on_camera_move(self, evt):
        x, y, z = evt.camera_location
        label = f"{x:.2f}, {y:.2f}, {z:.2f}"
        old_label = self._location_button.GetLabel()
        self._location_button.SetLabel(label)
        if len(label) != len(old_label):
            self.canvas.Layout()
        evt.Skip()

    def _on_speed_change(self, evt):
        label = f"{self.canvas.camera.move_speed*1000/33:.4g}{lang.get('program_3d_edit.file_ui.speed_blocks_per_second')}"
        old_label = self._speed_button.GetLabel()
        self._speed_button.SetLabel(label)
        if len(label) != len(old_label):
            self.canvas.Layout()
        evt.Skip()


class SpeedSelectDialog(wx.Dialog):
    def __init__(self, parent: wx.Window, speed: float):
        wx.Dialog.__init__(self, parent)
        self.SetTitle(lang.get("program_3d_edit.file_ui.speed_dialog_name"))

        sizer = wx.BoxSizer(wx.VERTICAL)

        self._speed_spin_ctrl_double = wx.SpinCtrlDouble(
            self, wx.ID_ANY, initial=speed, min=0.0, max=1_000_000_000.0
        )
        self._speed_spin_ctrl_double.SetToolTip(
            lang.get("program_3d_edit.file_ui.speed_tooltip")
        )

        def on_mouse_wheel(evt: wx.MouseEvent):
            if evt.GetWheelRotation() > 0:
                self._speed_spin_ctrl_double.SetValue(
                    self._speed_spin_ctrl_double.GetValue()
                    + self._speed_spin_ctrl_double.GetIncrement()
                )
            else:
                self._speed_spin_ctrl_double.SetValue(
                    self._speed_spin_ctrl_double.GetValue()
                    - self._speed_spin_ctrl_double.GetIncrement()
                )

        self._speed_spin_ctrl_double.Bind(wx.EVT_MOUSEWHEEL, on_mouse_wheel)
        self._speed_spin_ctrl_double.SetIncrement(1.0)
        self._speed_spin_ctrl_double.SetDigits(4)
        sizer.Add(
            self._speed_spin_ctrl_double, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5
        )

        button_sizer = wx.StdDialogButtonSizer()
        sizer.Add(button_sizer, 0, wx.ALIGN_RIGHT | wx.ALL, 4)

        self._button_ok = wx.Button(self, wx.ID_OK, "")
        self._button_ok.SetDefault()
        button_sizer.AddButton(self._button_ok)

        self._button_cancel = wx.Button(self, wx.ID_CANCEL, "")
        button_sizer.AddButton(self._button_cancel)

        button_sizer.Realize()

        self.SetSizer(sizer)
        sizer.Fit(self)

        self.SetAffirmativeId(self._button_ok.GetId())
        self.SetEscapeId(self._button_cancel.GetId())

        self.Layout()

    @property
    def speed(self) -> float:
        return self._speed_spin_ctrl_double.GetValue()



from typing import Optional
import logging
import sys

import wx
from wx import glcanvas
from OpenGL.GL import (
    GL_DEPTH_TEST,
    glEnable,
    GL_CULL_FACE,
    glDepthFunc,
    GL_LEQUAL,
    GL_BLEND,
    glBlendFunc,
    GL_SRC_ALPHA,
    GL_ONE_MINUS_SRC_ALPHA,
    glGetString,
    GL_VERSION,
)

log = logging.getLogger(__name__)


"""OpenGL workflow:
The initialisation function should be as minimal as possible. No OpenGL functions should be called here. The OpenGL state is not valid until the window is first shown.
You can implement functions that take a while in threads to not block the GUI but they must still not contain OpenGL functions.
Upon the window being shown the OpenGL context is activated and the state can be set in _init_opengl
Objects that need to bind textures or data should do so in the draw function so they can be sure the context is set.
"""


class BaseCanvas(glcanvas.GLCanvas):
    _context: Optional[glcanvas.GLContext]

    def __init__(self, parent: wx.Window):
        """
        Construct the canvas.
        No OpenGL interaction should be done here.
        OpenGL initialisation should be done in _init_opengl which is run after the window is first shown.
        """
        display_attributes = glcanvas.GLAttributes()
        display_attributes.PlatformDefaults().MinRGBA(8, 8, 8, 8).DoubleBuffer().Depth(
            24
        ).EndList()
        super().__init__(
            parent,
            display_attributes,
            size=parent.GetClientSize(),
            style=wx.WANTS_CHARS,
        )
        if sys.platform == "darwin":
            # This is required for MacOS. Amulet-Team/Amulet-Map-Editor#597
            context_attributes = wx.glcanvas.GLContextAttrs()
            context_attributes.CoreProfile().Robust().ResetIsolation().EndList()
            self._context = glcanvas.GLContext(
                self, ctxAttrs=context_attributes
            )  # setup the OpenGL context
        else:
            # This is required for linux and windows.
            # Amulet-Team/Amulet-Map-Editor#84
            # Amulet-Team/Amulet-Map-Editor#856
            self._context = glcanvas.GLContext(self)

        if not self._context.IsOK():
            raise Exception(f"Failed setting up context")

        self._init = False

        self.Bind(wx.EVT_SHOW, self._on_show)

    @property
    def context(self) -> glcanvas.GLContext:
        return self._context

    @property
    def context_identifier(self) -> str:
        # if not self._init:
        #     raise Exception("Cannot access the context until the window has been shown.")
        return str(id(self._context))

    def _on_show(self, evt: wx.ShowEvent):
        if not self._init and evt.IsShown():
            self._init = True
            self._init_opengl()

    def _init_opengl(self):
        """Set up the OpenGL state after the window is first shown."""
        self.SetCurrent(self._context)
        gl_version = glGetString(GL_VERSION)
        if isinstance(gl_version, bytes):
            gl_version = gl_version.decode("utf-8")
        log.info(f"OpenGL Version {gl_version}")
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_CULL_FACE)
        glDepthFunc(GL_LEQUAL)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)



from typing import TYPE_CHECKING
from amulet_map_editor.api.opengl.canvas_container import CanvasContainer

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class EditCanvasContainer(CanvasContainer):
    """A helper class to store a reference to the canvas.
    This ensures there are no memory leaks due to circular references."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__(canvas)

    @property
    def canvas(self) -> "EditCanvas":
        return self._canvas()



from __future__ import annotations

from typing import TYPE_CHECKING

from amulet.api.selection import SelectionGroup
from amulet.api.block import UniversalAirBlock
from amulet.api.data_types import Dimension, OperationReturnType

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def delete(
    world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
) -> OperationReturnType:
    internal_id = world.block_palette.get_add_block(UniversalAirBlock)

    iter_count = len(list(world.get_coord_box(dimension, selection, False)))
    count = 0

    for chunk, slices, _ in world.get_chunk_slice_box(dimension, selection, False):
        chunk.blocks[slices] = internal_id

        chunk_x, chunk_z = chunk.coordinates
        chunk_x *= 16
        chunk_z *= 16
        x_min = chunk_x + slices[0].start
        y_min = slices[1].start
        z_min = chunk_z + slices[2].start
        x_max = chunk_x + slices[0].stop
        y_max = slices[1].stop
        z_max = chunk_z + slices[2].stop

        for x, y, z in list(chunk.block_entities.keys()):
            if x_min <= x < x_max and y_min <= y < y_max and z_min <= z < z_max:
                chunk.block_entities.pop((x, y, z))

        chunk.changed = True
        count += 1
        yield count / iter_count



"""
This license applies to this file only.
-- begin license --
MIT License
Copyright (c) 2021 Amulet-Team
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
-- end license --
"""

from typing import TYPE_CHECKING, Tuple
import wx

from amulet.api.selection import SelectionGroup
from amulet.api.block import Block
from amulet.api.data_types import Dimension, OperationReturnType

from amulet_map_editor.api.wx.ui.base_select import EVT_PICK
from amulet_map_editor.api.wx.ui.block_select import BlockDefine
from amulet_map_editor.programs.edit.api.operations import DefaultOperationUI

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class Fill(wx.Panel, DefaultOperationUI):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        wx.Panel.__init__(self, parent)
        DefaultOperationUI.__init__(self, parent, canvas, world, options_path)
        self.Freeze()
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._sizer)

        options = self._load_options({})

        self._block_define = BlockDefine(
            self,
            world.translation_manager,
            wx.VERTICAL,
            *(options.get("fill_block_options", []) or [world.level_wrapper.platform]),
            show_pick_block=True
        )
        self._block_define.Bind(EVT_PICK, self._on_pick_block_button)
        self._sizer.Add(self._block_define, 1, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self._run_button = wx.Button(self, label="Run Operation")
        self._run_button.Bind(wx.EVT_BUTTON, self._run_operation)
        self._sizer.Add(self._run_button, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self.Layout()
        self.Thaw()

    @property
    def wx_add_options(self) -> Tuple[int, ...]:
        return (1,)

    def _on_pick_block_button(self, evt):
        """Set up listening for the block click"""
        self._show_pointer = True

    def _on_box_click(self):
        if self._show_pointer:
            self._show_pointer = False
            x, y, z = self._pointer.pointer_base
            self._block_define.universal_block = (
                self.world.get_block(x, y, z, self.canvas.dimension),
                None,
            )

    def _get_fill_block(self):
        return self._block_define.universal_block[0]

    def disable(self):
        self._save_options(
            {
                "fill_block": self._get_fill_block(),
                "fill_block_options": (
                    self._block_define.platform,
                    self._block_define.version_number,
                    self._block_define.force_blockstate,
                    self._block_define.namespace,
                    self._block_define.block_name,
                    self._block_define.properties,
                ),
            }
        )

    def _run_operation(self, _):
        self.canvas.run_operation(self._fill)

    def _fill(self):
        world = self.world
        dimension = self.canvas.dimension
        target_box = self.canvas.selection.selection_group
        fill_block, block_entity = self._block_define.universal_block

        internal_id = world.block_palette.get_add_block(fill_block)

        iter_count = len(list(world.get_coord_box(dimension, target_box, True)))
        count = 0

        for chunk, slices, _ in world.get_chunk_slice_box(dimension, target_box, True):
            chunk.blocks[slices] = internal_id

            chunk_x, chunk_z = chunk.coordinates
            chunk_x *= 16
            chunk_z *= 16
            x_min = chunk_x + slices[0].start
            y_min = slices[1].start
            z_min = chunk_z + slices[2].start
            x_max = chunk_x + slices[0].stop
            y_max = slices[1].stop
            z_max = chunk_z + slices[2].stop

            if block_entity is None:
                for x, y, z in list(chunk.block_entities.keys()):
                    if x_min <= x < x_max and y_min <= y < y_max and z_min <= z < z_max:
                        chunk.block_entities.pop((x, y, z))
            else:
                for x in range(x_min, x_max):
                    for y in range(y_min, y_max):
                        for z in range(z_min, z_max):
                            chunk.block_entities[(x, y, z)] = block_entity

            chunk.changed = True
            count += 1
            yield count / iter_count


export = {
    "name": "Fill",  # the name of the plugin
    "operation": Fill,  # the actual function to call when running the plugin
}



from __future__ import annotations
from typing import Type, Tuple

import atexit

import wx

from amulet_map_editor import CONFIG

PRE_EXISTING_CONFIG = CONFIG.get("window_preferences", {})


def write_config():
    CONFIG.put("window_preferences", PRE_EXISTING_CONFIG)


atexit.register(write_config)


class ExtendedTLW(wx.TopLevelWindow):
    __resized: bool
    __size: Tuple[int, int]
    __position: Tuple[int, int]


def on_idle(self: ExtendedTLW):
    qualified_name = ".".join((self.__module__, self.__class__.__name__))

    def wrapper(evt):
        if self.__resized:
            self.__resized = False
            PRE_EXISTING_CONFIG[qualified_name] = {
                "size": self.__size,
                "position": self.__position,
                "is_full_screen": self.IsMaximized(),
            }
        evt.Skip()

    return wrapper


def on_resize(self: ExtendedTLW):
    def wrapper(evt):
        self.__resized = True
        if not self.IsMaximized():
            # only store the non-maximised state
            self.__position = self.GetPosition().Get()
            self.__size = self.GetSize().Get()
        evt.Skip()

    return wrapper


def preserve_ui_preferences(clazz: Type[wx.TopLevelWindow]):
    assert issubclass(
        clazz, wx.TopLevelWindow
    ), "This takes a subclass of a top level window."
    original_init = clazz.__init__
    qualified_name = ".".join((clazz.__module__, clazz.__name__))

    def __init__(self: wx.TopLevelWindow, *args, **kwargs):
        original_init(self, *args, **kwargs)
        self.__resized = False
        self.__position = self.GetPosition().Get()
        self.__size = self.GetSize().Get()
        self: ExtendedTLW

        if qualified_name in PRE_EXISTING_CONFIG:
            window_config = PRE_EXISTING_CONFIG[qualified_name]
            self.SetPosition(window_config["position"])
            self.SetSize(window_config["size"])
            self.Maximize(window_config.get("is_full_screen", False))
        else:
            self.Maximize()
        self.Layout()
        self.Refresh()

        self.Bind(wx.EVT_MOVE, on_resize(self))
        self.Bind(wx.EVT_SIZE, on_resize(self))
        self.Bind(wx.EVT_IDLE, on_idle(self))

    clazz.__init__ = __init__
    return clazz



#!/usr/bin/env python3


def _on_error(e):
    """Code to handle errors"""
    try:
        import traceback
        import sys
        import os

    except ImportError as e:
        # Something has gone seriously wrong
        print(e)
        print("Failed to import requirements. Check that you extracted correctly.")
        input("Press ENTER to continue.")
    else:
        err = "\n".join(
            [traceback.format_exc()]
            + ["Failed to import requirements. Check that you extracted correctly."]
            * isinstance(e, ImportError)
            + [str(e)]
        )
        print(err)
        try:
            with open("crash.log", "w") as f:
                f.write(err)
        except OSError:
            pass
        input("Press ENTER to continue.")
        sys.exit(1)


try:
    import sys

    if sys.version_info[:2] < (3, 7):
        raise Exception("Must be using Python 3.7+")
    import logging
    import os
    import traceback
    import glob
    import time
    import wx
    import platformdirs

    if sys.platform == "linux" and wx.VERSION >= (4, 1, 1):
        # bug 247
        os.environ["PYOPENGL_PLATFORM"] = "egl"
except Exception as e_:
    _on_error(e_)


def _init_log():
    logs_path = os.environ["LOG_DIR"]
    # set up handlers
    os.makedirs(logs_path, exist_ok=True)
    # remove all log files older than a week
    for path in glob.glob(os.path.join(glob.escape(logs_path), "*.log")):
        if (
            os.path.isfile(path)
            and os.path.getmtime(path) < time.time() - 3600 * 24 * 7
        ):
            os.remove(path)

    log = logging.getLogger()
    log.setLevel(logging.DEBUG if "amulet-debug" in sys.argv else logging.INFO)

    file_handler = logging.FileHandler(
        os.path.join(logs_path, f"amulet_{os.getpid()}.log"), "w", encoding="utf-8"
    )
    file_handler.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    log.addHandler(file_handler)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter("%(levelname)s - %(message)s"))
    log.addHandler(console_handler)


def main():
    try:
        # Initialise default paths.
        data_dir = platformdirs.user_data_dir("AmuletMapEditor", "AmuletTeam")
        os.environ.setdefault("DATA_DIR", data_dir)
        config_dir = platformdirs.user_config_dir("AmuletMapEditor", "AmuletTeam")
        if config_dir == data_dir:
            config_dir = os.path.join(data_dir, "Config")
        os.environ.setdefault("CONFIG_DIR", config_dir)
        os.environ.setdefault(
            "CACHE_DIR", platformdirs.user_cache_dir("AmuletMapEditor", "AmuletTeam")
        )
        os.environ.setdefault(
            "LOG_DIR", platformdirs.user_log_dir("AmuletMapEditor", "AmuletTeam")
        )

        _init_log()
        from amulet_map_editor.api.framework import AmuletApp

    except Exception as e:
        _on_error(e)
    else:
        try:
            app = AmuletApp(0)
            app.MainLoop()
        except Exception as e:
            log = logging.getLogger(__name__)
            log.critical(
                f"Amulet Crashed. Sorry about that. Please report it to a developer if you think this is an issue. \n{traceback.format_exc()}"
            )
            input("Press ENTER to continue.")

    sys.exit(0)


if __name__ == "__main__":
    main()



class BaseTab:
    def enable(self):
        """Run when the tab is shown/enabled"""
        pass

    def can_disable(self) -> bool:
        """
        Check if it is safe to disable the tab.
        :return: True if the tab can be disabled, False otherwise.
        """
        return True

    def disable(self):
        """Run when the tab is hidden/disabled"""
        pass

    def can_close(self) -> bool:
        """
        Check if it is safe to close the tab.
        If this is going to return False it should notify the user.
        :return: True if the tab can be closed, False otherwise.
        """
        return True

    def close(self):
        """Fully close the tab. Called when destroying the tab."""
        pass



"""
This license applies to this file only.
-- begin license --
MIT License
Copyright (c) 2021 Amulet-Team
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
-- end license --
"""

from typing import TYPE_CHECKING

import wx
import math

from amulet.utils import block_coords_to_chunk_coords
from amulet.api.chunk.biomes import BiomesShape
from amulet_map_editor.api.wx.ui.base_select import EVT_PICK
from amulet_map_editor.api.wx.ui.biome_select import BiomeDefine
from amulet_map_editor.programs.edit.api.operations import SimpleOperationPanel
from amulet_map_editor.api.wx.ui.simple import SimpleChoiceAny

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas
    from amulet.api.selection import SelectionGroup
    from amulet.api.data_types import Dimension, OperationReturnType


BoxMode = "box"
ColumnMode = "column"

lang = {
    ColumnMode: "Selected Column",
    BoxMode: "Selected Box",
}


MODES = {
    ColumnMode: "Set the biomes for the vertical column the selection intersects.",
    BoxMode: "Set the biomes for only the blocks in the selection.",
}

Border = wx.TOP | wx.LEFT | wx.RIGHT | wx.EXPAND


class SetBiome(SimpleOperationPanel):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        SimpleOperationPanel.__init__(self, parent, canvas, world, options_path)
        self.Freeze()
        options = self._load_options({})

        self._mode = SimpleChoiceAny(self, sort=False)
        self._mode.SetItems({mode: lang[mode] for mode in MODES.keys()})
        self._sizer.Add(self._mode, 0, Border, 5)
        self._mode.Bind(wx.EVT_CHOICE, self._on_mode_change)

        self._mode_description = wx.TextCtrl(
            self, style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_BESTWRAP
        )
        self._sizer.Add(self._mode_description, 0, Border, 5)

        self._mode_description.SetLabel(MODES[self._mode.GetCurrentObject()])
        self._mode_description.Fit()

        self._biome_choice = BiomeDefine(
            self,
            world.level_wrapper.translation_manager,
            wx.VERTICAL,
            *(
                options.get("original_block_options", [])
                or [world.level_wrapper.platform]
            ),
            show_pick_biome=True,
        )
        self._biome_choice.Bind(EVT_PICK, self._on_pick_biome_button)
        self._sizer.Add(self._biome_choice, 1, Border, 5)

        self._add_run_button()

        self.Thaw()

    def _on_mode_change(self, evt):
        self._mode_description.SetLabel(MODES[self._mode.GetCurrentObject()])
        self._mode_description.Fit()
        self.Layout()
        evt.Skip()

    def _on_pick_biome_button(self, evt):
        """Set up listening for the biome click"""
        self._show_pointer = True

    def _on_box_click(self):
        if self._show_pointer:
            self._show_pointer = False
            x, y, z = self._pointer.pointer_base

            # TODO: replace with "get_biome(x, y, z)" if it'll be created
            cx, cz = block_coords_to_chunk_coords(
                x, z, sub_chunk_size=self.world.sub_chunk_size
            )
            offset_x, offset_z = x - 16 * cx, z - 16 * cz
            chunk = self.world.get_chunk(cx, cz, self.canvas.dimension)

            if chunk.biomes.dimension == BiomesShape.Shape3D:
                biome = chunk.biomes[offset_x // 4, y // 4, offset_z // 4]
            elif chunk.biomes.dimension == BiomesShape.Shape2D:
                biome = chunk.biomes[offset_x, offset_z]
            else:
                return

            self._biome_choice.universal_biome = chunk.biome_palette[biome]

    def _operation(
        self, world: "BaseLevel", dimension: "Dimension", selection: "SelectionGroup"
    ) -> "OperationReturnType":
        mode = self._mode.GetCurrentObject()

        iter_count = len(list(world.get_coord_box(dimension, selection, False)))
        for count, (chunk, slices, _) in enumerate(
            world.get_chunk_slice_box(dimension, selection, False)
        ):
            new_biome = chunk.biome_palette.get_add_biome(
                self._biome_choice.universal_biome
            )

            if mode == BoxMode:
                if chunk.biomes.dimension == BiomesShape.Shape3D:
                    slices = (
                        slice(slices[0].start // 4, math.ceil(slices[0].stop / 4)),
                        slice(slices[1].start // 4, math.ceil(slices[1].stop / 4)),
                        slice(slices[2].start // 4, math.ceil(slices[2].stop / 4)),
                    )
                elif chunk.biomes.dimension == BiomesShape.Shape2D:
                    slices = (slices[0], slices[2])
                else:
                    continue
            elif mode == ColumnMode:
                if chunk.biomes.dimension == BiomesShape.Shape3D:
                    bounds = world.bounds(dimension)
                    slices = (
                        slice(slices[0].start // 4, math.ceil(slices[0].stop / 4)),
                        # TODO this might cause issues with infinite height worlds. The biome handling needs a rewrite
                        slice(bounds.min_y // 4, math.ceil(bounds.max_y / 4)),
                        slice(slices[2].start // 4, math.ceil(slices[2].stop / 4)),
                    )
                elif chunk.biomes.dimension == BiomesShape.Shape2D:
                    slices = (slices[0], slices[2])
                else:
                    continue
            else:
                raise ValueError(
                    f"mode {mode} is not a valid mode for the Set Biome operation."
                )

            chunk.biomes[slices] = new_biome

            chunk.changed = True
            yield (count + 1) / iter_count


export = {
    "name": "Set Biome",  # the name of the plugin
    "operation": SetBiome,  # the actual function to call when running the plugin
}



import sys
from typing import Tuple, Dict, Type
from types import ModuleType
import os
import traceback
import pkgutil
import importlib
import logging

from .loader import (
    BaseOperationLoader,
    OperationLoader,
    UIOperationLoader,
)
from .util import STOCK_PLUGINS_DIR, STOCK_PLUGINS_NAME, CUSTOM_PLUGINS_DIR

log = logging.getLogger(__name__)


class BaseOperationManager:
    OperationClass: Type[BaseOperationLoader] = None

    def __init__(self, group_name: str):
        """Set up an operation manager that handles operations for a set group.

        :param group_name: The name of the group. This is used to pick the directory.
        """
        # The loaded operations. Stored under their identifier.
        assert issubclass(self.OperationClass, BaseOperationLoader)
        self._group_name = group_name
        self._operations: Dict[str, BaseOperationLoader] = {}
        self.reload()

    @property
    def public_path(self) -> str:
        return os.path.abspath(os.path.join(CUSTOM_PLUGINS_DIR, self._group_name))

    @property
    def operations(self) -> Tuple[BaseOperationLoader, ...]:
        return tuple(self._operations.values())

    def get_operation(self, operation_id: str):
        return self._operations[operation_id]

    def __getitem__(self, operation_id: str):
        return self._operations[operation_id]

    def reload(self):
        """Unload the old operations and reload them all.
        If new operations are created, load them.
        If old operations were removed, remove them."""
        self._operations.clear()
        self._load_internal_submodules(
            os.path.join(STOCK_PLUGINS_DIR, self._group_name),
            ".".join([STOCK_PLUGINS_NAME, self._group_name]),
        )
        self._load_external_submodules(self.public_path)

    def _load_internal_submodules(self, package_path: str, package_name: str):
        """
        Load submodules from within a known package.

        :param package_path: The file path of the package to load submodules from.
        :param package_name: The python path of the package to load submodules from.
        """
        self._load_submodules(package_path, f"{package_name}.")

    def _load_external_submodules(self, path: str):
        """
        Load submodules from an external package.

        Note that the path will be added to the system path.

        :param path: The path to the directory to find modules in.
        """

        # create the directory
        os.makedirs(path, exist_ok=True)

        if path not in sys.path:
            # The path modules are loaded from should be on the system path
            sys.path.append(path)

        self._load_submodules(path)

    def _load_submodules(self, path: str, package: str = ""):
        """
        Load all operations in a path.
        You should not use this directly.
        Use _load_internal_submodules or _load_external_submodules

        :param path: The path to load modules from.
        :param package: The package name prefix
        """
        # clean the path
        path = os.path.abspath(path)

        for _, module_name, _ in pkgutil.iter_modules([path]):
            module_name = f"{package}{module_name}"
            try:
                mod = importlib.import_module(module_name)
                mod = importlib.reload(mod)
            except BaseException:
                log.warning(
                    f"There was an exception while loading operation {module_name}.\n{traceback.format_exc()}"
                )
            else:
                if (
                    os.path.dirname(
                        mod.__path__[0] if hasattr(mod, "__path__") else mod.__file__
                    )
                    == path
                ):
                    # If the loaded module is actually in the path it was loaded from.
                    # There may be cases where the name is shadowed by another module.
                    self._load_module(module_name, mod)
                else:
                    log.warning(
                        f"Module {module_name} shadows another module. Try using a different module name."
                    )

    def _load_module(self, obj_path: str, mod: ModuleType):
        if hasattr(mod, "export"):
            export = mod.export
            if isinstance(export, dict):
                self._load_operation(obj_path, export)
            elif isinstance(export, (list, tuple)):
                for i, export_dict in export:
                    self._load_operation(f"{obj_path}[{i}]", export_dict)
            else:
                log.error(f"The format of export in {obj_path} is invalid.")
        else:
            log.error(f"export is not present in {obj_path}")

    def _load_operation(self, identifier: str, export_dict: dict):
        """Create an instance of a subclass of BaseOperationLoader to load the operation.
        When finished and the class is valid store it in self._operations

        :param identifier: A unique identifier for this operation. <path>[0]
        :param export_dict: The dictionary defining the operation.
        :return:
        """
        op = self.OperationClass(identifier, export_dict)
        if op.is_valid:
            self._operations[op.identifier] = op


class OperationManager(BaseOperationManager):
    OperationClass = OperationLoader
    """A class to manage a group of function operations that run immediately when called."""


class UIOperationManager(BaseOperationManager):
    OperationClass = UIOperationLoader
    """A class to manage a group of UI operations that return a wx UI when called."""



import webbrowser
import wx
import wx.adv
import wx.lib.inspection

from amulet_map_editor.api import image, lang
from .base_page import BasePageUI
from amulet_map_editor.api.wx.ui.select_world import open_level_from_dialog


class AmuletMainMenu(wx.Panel, BasePageUI):
    def __init__(self, parent: wx.Window):
        super(AmuletMainMenu, self).__init__(parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        sizer.AddStretchSpacer(1)
        name_sizer = wx.BoxSizer()
        sizer.Add(name_sizer, 0, wx.CENTER)
        icon_img = image.logo.amulet_logo.bitmap(64, 64)

        icon = wx.StaticBitmap(self, wx.ID_ANY, icon_img, (0, 0), (64, 64))
        icon2 = wx.StaticBitmap(self, wx.ID_ANY, icon_img, (0, 0), (64, 64))
        icon2.Bind(
            wx.EVT_LEFT_DOWN, lambda evt: wx.lib.inspection.InspectionTool().Show()
        )
        name_sizer.Add(icon, flag=wx.CENTER)

        self._amulet_name = wx.StaticText(self)
        self._amulet_name.SetFont(wx.Font(40, wx.DECORATIVE, wx.NORMAL, wx.NORMAL))
        name_sizer.Add(
            self._amulet_name, flag=wx.CENTER | wx.LEFT | wx.RIGHT, border=10
        )
        name_sizer.Add(icon2, flag=wx.CENTER)
        button_font = wx.Font(20, wx.DECORATIVE, wx.NORMAL, wx.NORMAL)
        self._open_world_button = wx.Button(self, size=(400, 70))
        self._open_world_button.SetFont(button_font)
        self._open_world_button.Bind(
            wx.EVT_BUTTON, lambda _: open_level_from_dialog(self)
        )
        sizer.Add(self._open_world_button, 0, wx.ALL | wx.CENTER, 5)

        self._user_manual_button = wx.Button(self, size=(400, 70))
        self._user_manual_button.SetFont(button_font)
        self._user_manual_button.Bind(wx.EVT_BUTTON, self._documentation)
        sizer.Add(self._user_manual_button, 0, wx.ALL | wx.CENTER, 5)

        self._bug_tracker_button = wx.Button(self, size=(400, 70))
        self._bug_tracker_button.SetFont(button_font)
        self._bug_tracker_button.Bind(wx.EVT_BUTTON, self._bugs)
        sizer.Add(self._bug_tracker_button, 0, wx.ALL | wx.CENTER, 5)

        self._discord_button = wx.Button(self, size=(400, 70))
        self._discord_button.SetFont(button_font)
        self._discord_button.Bind(wx.EVT_BUTTON, self._discord)
        sizer.Add(self._discord_button, 0, wx.ALL | wx.CENTER, 5)

        sizer.AddStretchSpacer(2)

        self._lang_button = wx.BitmapButton(
            self, bitmap=image.icon.tablericons.language.bitmap(64, 64)
        )
        self._lang_button.Bind(wx.EVT_BUTTON, self._select_language)
        sizer.Add(self._lang_button, 0, wx.ALIGN_RIGHT)

        self._load_strings()

    def _load_strings(self):
        self._amulet_name.SetLabel(lang.get("meta.amulet"))
        self._open_world_button.SetLabel(lang.get("main_menu.open_world"))
        self._user_manual_button.SetLabel(lang.get("main_menu.user_manual"))
        self._user_manual_button.SetToolTip(lang.get("app.browser_open_tooltip"))
        self._bug_tracker_button.SetLabel(lang.get("main_menu.bug_tracker"))
        self._bug_tracker_button.SetToolTip(lang.get("app.browser_open_tooltip"))
        self._discord_button.SetLabel(lang.get("main_menu.discord"))
        self._discord_button.SetToolTip(lang.get("app.browser_open_tooltip"))

    @staticmethod
    def _documentation(_):
        webbrowser.open(
            "https://github.com/Amulet-Team/Amulet-Map-Editor/blob/master/amulet_map_editor/readme.md"
        )

    @staticmethod
    def _bugs(_):
        webbrowser.open(
            "https://github.com/Amulet-Team/Amulet-Map-Editor/issues?q=is%3Aissue"
        )

    @staticmethod
    def _discord(_):
        webbrowser.open("https://www.amuletmc.com/discord")

    def enable(self):
        self.GetTopLevelParent().create_menu()

    def _select_language(self, evt):
        dialog = LangSelectDialog(self)
        if dialog.ShowModal() == wx.ID_OK:
            lang.set_language(dialog.get_language())
        dialog.Destroy()
        self._load_strings()


class LangSelectDialog(wx.Dialog):
    def __init__(self, *args, **kwds):
        # begin wxGlade: LangSelectDialog.__init__
        kwds["style"] = (
            kwds.get("style", 0) | wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER
        )
        wx.Dialog.__init__(self, *args, **kwds)
        self.SetTitle(lang.get("language_select.title"))

        sizer_1 = wx.BoxSizer(wx.VERTICAL)

        self._label = wx.StaticText(self, label=lang.get("language_select.help"))
        sizer_1.Add(self._label, 0, wx.ALIGN_CENTER)

        self.hyperlink_1 = wx.adv.HyperlinkCtrl(
            self,
            wx.ID_ANY,
            lang.get("language_select.contribute"),
            "https://github.com/Amulet-Team/Amulet-Map-Editor#contributing",
        )
        sizer_1.Add(self.hyperlink_1, 0, wx.ALIGN_CENTER)

        self._lang_list_box = wx.ListBox(self, choices=lang.get_languages())
        self._lang_list_box.SetSelection(
            self._lang_list_box.FindString(lang.get_language())
        )
        sizer_1.Add(self._lang_list_box, 1, wx.EXPAND, 0)

        sizer_2 = wx.StdDialogButtonSizer()
        sizer_1.Add(sizer_2, 0, wx.ALIGN_RIGHT | wx.ALL, 4)

        self._button_ok = wx.Button(self, wx.ID_OK, "")
        self._button_ok.SetDefault()
        sizer_2.AddButton(self._button_ok)

        self._button_cancel = wx.Button(self, wx.ID_CANCEL, "")
        sizer_2.AddButton(self._button_cancel)

        sizer_2.Realize()

        self.SetSizer(sizer_1)
        sizer_1.Fit(self)

        self.SetAffirmativeId(self._button_ok.GetId())
        self.SetEscapeId(self._button_cancel.GetId())

        self.Layout()

    def get_language(self):
        return self._lang_list_box.GetStringSelection()



import os
import pickle
import gzip
from typing import Any

_path = os.path.abspath(os.path.join(os.environ["CONFIG_DIR"]))


def get(identifier: str, default: Any = None) -> Any:
    """
    Get the config data for a given identifier. Use an identifier unique to your program.
    :param identifier: An identifier unique to your program
    :param default: The value to return if one could not be loaded
    :return: config data
    """
    path = os.path.join(_path, identifier + ".config")
    if os.path.isfile(path):
        try:
            with gzip.open(path, "rb") as fp:
                val = pickle.load(fp)
            return val
        except:
            pass
    return default


def put(identifier: str, data: Any):
    """
    Add data to the config file and save to disk.
    :param identifier: An identifier unique to your program
    :param data: The data to be saved. Must be JSON serialisable
    :return:
    """
    if not os.path.isdir(_path):
        os.makedirs(_path)
    path = os.path.join(_path, identifier + ".config")
    with gzip.open(path, "wb") as fp:
        pickle.dump(data, fp)



import wx
from wx.lib import newevent
from typing import Tuple, Dict, Optional, List, Union
import weakref

import PyMCTranslate
import amulet_nbt
from amulet_nbt import SNBTType
from amulet.api.block import PropertyDataTypes, PropertyType

WildcardSNBTType = Union[SNBTType, str]

from amulet_map_editor.api.image import ADD_ICON, SUBTRACT_ICON

(
    PropertiesChangeEvent,
    EVT_PROPERTIES_CHANGE,
) = newevent.NewCommandEvent()  # the properties changed


class PropertySelect(wx.Panel):
    def __init__(
        self,
        parent: wx.Window,
        translation_manager: PyMCTranslate.TranslationManager,
        platform: str,
        version_number: Tuple[int, int, int],
        force_blockstate: bool,
        namespace: str,
        block_name: str,
        properties: Dict[str, SNBTType] = None,
        wildcard_mode=False,
    ):
        super().__init__(parent, style=wx.BORDER_SIMPLE)
        self._parent = weakref.ref(parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        self._translation_manager = translation_manager

        self._platform: Optional[str] = None
        self._version_number: Optional[Tuple[int, int, int]] = None
        self._force_blockstate: Optional[bool] = None
        self._namespace: Optional[str] = None
        self._block_name: Optional[str] = None

        self._manual_enabled = False
        self._simple = SimplePropertySelect(self, translation_manager, wildcard_mode)
        sizer.Add(self._simple, 1, wx.EXPAND)
        self._manual = ManualPropertySelect(self, translation_manager)
        sizer.Add(self._manual, 1, wx.EXPAND)

        self._wildcard_mode = wildcard_mode

        self._set_version_block(
            (platform, version_number, force_blockstate, namespace, block_name)
        )
        self.set_properties(properties)

    @property
    def parent(self) -> wx.Window:
        return self._parent()

    @property
    def wildcard_mode(self) -> bool:
        return self._wildcard_mode

    @property
    def version_block(self) -> Tuple[str, Tuple[int, int, int], bool, str, str]:
        return (
            self._platform,
            self._version_number,
            self._force_blockstate,
            self._namespace,
            self._block_name,
        )

    @version_block.setter
    def version_block(
        self, version_block: Tuple[str, Tuple[int, int, int], bool, str, str]
    ):
        self._set_version_block(version_block)
        self.str_properties = None

    def _set_version_block(
        self, version_block: Tuple[str, Tuple[int, int, int], bool, str, str]
    ):
        version = version_block[:3]
        assert (
            version[0] in self._translation_manager.platforms()
            and version[1] in self._translation_manager.version_numbers(version[0])
            and isinstance(version[2], bool)
        ), f"{version} is not a valid version"
        self._platform, self._version_number, self._force_blockstate = version
        block = version_block[3:5]
        assert isinstance(block[0], str) and isinstance(
            block[1], str
        ), "The block namespace and block name must be strings"
        self._namespace, self._block_name = block
        self._set_ui()

    @property
    def str_properties(self) -> Dict[str, WildcardSNBTType]:
        if self._manual_enabled:
            return self._manual.properties
        else:
            return self._simple.properties

    @str_properties.setter
    def str_properties(self, properties: Dict[str, WildcardSNBTType]):
        self.set_properties(properties)
        wx.PostEvent(
            self, PropertiesChangeEvent(self.GetId(), properties=self.str_properties)
        )

    @property
    def properties(self) -> PropertyType:
        if self.wildcard_mode:
            raise Exception(
                "Accessing the properties attribute is invalid in wildcard mode"
            )
        return {
            key: amulet_nbt.from_snbt(val) for key, val in self.str_properties.items()
        }

    @properties.setter
    def properties(self, properties: PropertyType):
        self.str_properties = {
            key: val.to_snbt() for key, val in (properties or {}).items()
        }

    def set_properties(self, properties: Dict[str, SNBTType]):
        properties = properties or {}
        self.Freeze()
        if self._manual_enabled:
            self._manual.properties = properties
        else:
            self._simple.properties = properties
        self.TopLevelParent.Layout()
        self.Thaw()

    def _set_ui(self):
        self.Freeze()
        translator = self._translation_manager.get_version(
            self._platform, self._version_number
        ).block

        self._manual_enabled = self._block_name not in translator.base_names(
            self._namespace, self._force_blockstate
        )
        if self._manual_enabled:
            self._simple.Hide()
            self._manual.Show()
        else:
            self._simple.Show()
            self._simple.set_specification(
                translator.get_specification(
                    self._namespace, self._block_name, self._force_blockstate
                )
            )
            self._manual.Hide()
        self.Thaw()


class SimplePropertySelect(wx.Panel):
    def __init__(
        self,
        parent: wx.Window,
        translation_manager: PyMCTranslate.TranslationManager,
        wildcard_mode,
    ):
        super().__init__(parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self._translation_manager = translation_manager

        header_sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(header_sizer, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5)
        label = wx.StaticText(self, label="Property Name", style=wx.ALIGN_CENTER)
        header_sizer.Add(label, 1)
        label = wx.StaticText(
            self, label="Property Value (SNBT)", style=wx.ALIGN_CENTER
        )
        header_sizer.Add(label, 1, wx.LEFT, 5)
        self._property_sizer = wx.GridSizer(2, 5, 5)
        sizer.Add(self._property_sizer, 0, wx.ALL | wx.EXPAND, 5)

        self._properties: Dict[str, wx.Choice] = {}
        self._specification: dict = {}
        self._wildcard_mode = wildcard_mode

    def set_specification(self, specification: dict):
        self._specification = specification

    @property
    def properties(self) -> Dict[str, SNBTType]:
        return {
            name: choice.GetString(choice.GetSelection())
            for name, choice in self._properties.items()
        }

    @properties.setter
    def properties(self, properties: Dict[str, SNBTType]):
        self.Freeze()
        self._properties.clear()
        self._property_sizer.Clear(True)
        spec_properties: Dict[str, List[str]] = self._specification.get(
            "properties", {}
        )
        spec_defaults = self._specification.get("defaults", {})

        for name, choices in spec_properties.items():
            label = wx.StaticText(self, label=name)
            self._property_sizer.Add(label, 0, wx.ALIGN_CENTER)
            if self._wildcard_mode:
                choices = ["*"] + choices
            choice = wx.Choice(self, choices=choices)
            self._property_sizer.Add(choice, 0, wx.EXPAND)
            choice.Bind(
                wx.EVT_CHOICE,
                lambda evt: wx.PostEvent(
                    self,
                    PropertiesChangeEvent(self.GetId(), properties=self.properties),
                ),
            )
            val = spec_defaults[name]
            if name in properties and val != "*":
                try:
                    snbt = amulet_nbt.from_snbt(properties[name]).to_snbt()
                except:
                    pass
                else:
                    if snbt in choices:
                        val = snbt
            if val in choices:
                choice.SetSelection(choices.index(val))
            self._properties[name] = choice
        self.Thaw()
        self.Fit()
        self.Layout()


class ManualPropertySelect(wx.Panel):
    def __init__(
        self, parent: wx.Window, translation_manager: PyMCTranslate.TranslationManager
    ):
        super().__init__(parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        self._translation_manager = translation_manager

        header_sizer = wx.BoxSizer(wx.HORIZONTAL)
        add_button = wx.BitmapButton(
            self, bitmap=ADD_ICON.bitmap(30, 30), size=(30, 30)
        )
        header_sizer.Add(add_button)
        sizer.Add(header_sizer, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5)
        label = wx.StaticText(self, label="Property Name", style=wx.ALIGN_CENTER)
        header_sizer.Add(label, 1, wx.LEFT | wx.ALIGN_CENTER_VERTICAL, 5)
        label = wx.StaticText(
            self, label="Property Value (SNBT)", style=wx.ALIGN_CENTER
        )
        header_sizer.Add(label, 1, wx.LEFT | wx.ALIGN_CENTER_VERTICAL, 5)
        header_sizer.AddStretchSpacer(1)

        self._property_sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(
            self._property_sizer, 0, wx.LEFT | wx.RIGHT | wx.BOTTOM | wx.EXPAND, 5
        )

        add_button.Bind(wx.EVT_BUTTON, lambda evt: self._add_property())

        self._property_index = 0
        self._properties: Dict[int, Tuple[wx.TextCtrl, wx.TextCtrl]] = {}

    def _post_property_change(self):
        wx.PostEvent(
            self, PropertiesChangeEvent(self.GetId(), properties=self.properties)
        )

    def _add_property(self, name: str = "", value: SNBTType = ""):
        self.Freeze()
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._property_index += 1
        subtract_button = wx.BitmapButton(
            self, bitmap=SUBTRACT_ICON.bitmap(30, 30), size=(30, 30)
        )
        sizer.Add(subtract_button, 0, wx.ALIGN_CENTER_VERTICAL)
        index = self._property_index
        subtract_button.Bind(
            wx.EVT_BUTTON, lambda evt: self._on_remove_property(sizer, index)
        )
        name_entry = wx.TextCtrl(self, value=name, style=wx.TE_CENTER)
        sizer.Add(name_entry, 1, wx.ALIGN_CENTER_VERTICAL | wx.LEFT, 5)
        name_entry.Bind(wx.EVT_TEXT, lambda evt: self._post_property_change())
        value_entry = wx.TextCtrl(self, value=value, style=wx.TE_CENTER)
        sizer.Add(value_entry, 1, wx.ALIGN_CENTER_VERTICAL | wx.LEFT, 5)
        snbt_text = wx.StaticText(self, style=wx.ALIGN_CENTER)
        sizer.Add(snbt_text, 1, wx.ALIGN_CENTER_VERTICAL | wx.LEFT, 5)
        self._change_value("", snbt_text)
        value_entry.Bind(wx.EVT_TEXT, lambda evt: self._on_value_change(evt, snbt_text))

        self._property_sizer.Add(sizer, 1, wx.TOP | wx.EXPAND, 5)
        self._properties[self._property_index] = (name_entry, value_entry)
        self.Fit()
        self.TopLevelParent.Layout()
        self.Thaw()

    def _on_value_change(self, evt, snbt_text: wx.StaticText):
        self._change_value(evt.GetString(), snbt_text)
        self._post_property_change()
        evt.Skip()

    def _change_value(self, snbt: SNBTType, snbt_text: wx.StaticText):
        try:
            nbt = amulet_nbt.from_snbt(snbt)
        except:
            snbt_text.SetLabel("Invalid SNBT")
            snbt_text.SetBackgroundColour((255, 200, 200))
        else:
            if isinstance(nbt, PropertyDataTypes):
                snbt_text.SetLabel(nbt.to_snbt())
                snbt_text.SetBackgroundColour(wx.NullColour)
            else:
                snbt_text.SetLabel(f"{nbt.__class__.__name__} not valid")
                snbt_text.SetBackgroundColour((255, 200, 200))
        self.Layout()

    def _on_remove_property(self, sizer: wx.Sizer, key: int):
        self.Freeze()
        self._property_sizer.Detach(sizer)
        sizer.Clear(True)
        del self._properties[key]
        self.TopLevelParent.Layout()
        self.Thaw()
        self._post_property_change()

    @property
    def properties(self) -> Dict[str, SNBTType]:
        properties = {}
        for name, value in self._properties.values():
            try:
                nbt = amulet_nbt.from_snbt(value.GetValue())
            except:
                continue
            if name.GetValue() and isinstance(nbt, PropertyDataTypes):
                properties[name.GetValue()] = nbt.to_snbt()
        return properties

    @properties.setter
    def properties(self, properties: Dict[str, SNBTType]):
        self._property_sizer.Clear(True)
        self._properties.clear()
        self._property_index = 0
        for name, value in properties.items():
            self._add_property(name, value)


if __name__ == "__main__":

    def main():
        translation_manager = PyMCTranslate.new_translation_manager()
        app = wx.App()
        dialog = wx.Dialog(None)
        sizer = wx.BoxSizer()
        dialog.SetSizer(sizer)
        sizer.Add(
            PropertySelect(
                dialog,
                translation_manager,
                "java",
                (1, 16, 0),
                False,
                "minecraft",
                "oak_fence",
            ),
            1,
            wx.ALL,
            5,
        )
        dialog.Show()
        dialog.Fit()
        app.MainLoop()

    main()



from typing import Union
import numpy

from .render_selection import RenderSelection
from amulet_map_editor.api.opengl.resource_pack import OpenGLResourcePack
from amulet_map_editor.api.opengl.data_types import RGBColour
from .colours import colours


class RenderSelectionHighlightable(RenderSelection):
    """A drawable selection box with edges that can be highlighted"""

    def __init__(self, context_identifier: str, resource_pack: OpenGLResourcePack):
        super().__init__(context_identifier, resource_pack)
        # which edges are highlighted
        self._highlight_edges = numpy.zeros((2, 3), dtype=bool)

    @property
    def highlight_colour(self) -> RGBColour:
        return colours.get("box_highlight", (0.5, 0.5, 1.0))

    def reset_highlight_edges(self):
        if numpy.any(self._highlight_edges):
            self.set_highlight_edges(False)

    def set_highlight_edges(self, highlight_edges: Union[numpy.ndarray, bool]):
        self._highlight_edges[:] = highlight_edges
        self._mark_recreate()

    def _create_geometry_(self):
        super()._create_geometry_()
        self.verts[:36, 9:12] = self.box_tint

        indexes = numpy.zeros(6, numpy.uint8)
        if self.point2[0] > self.point1[0]:
            indexes[[1, 3]] = 0, 3
        else:
            indexes[[1, 3]] = 3, 0

        if self.point2[1] > self.point1[1]:
            indexes[[0, 5]] = 1, 4
        else:
            indexes[[0, 5]] = 4, 1

        if self.point2[2] > self.point1[2]:
            indexes[[2, 4]] = 2, 5
        else:
            indexes[[2, 4]] = 5, 2

        # [1, 0, 2, 3, 5, 4]

        # 0 down 1
        # 1 west 0
        # 2 north 2
        # 3 east 3
        # 4 south 5
        # 5 up 4

        self.verts[:36][
            numpy.repeat(self._highlight_edges.ravel()[indexes], 6), 9:12
        ] = self.highlight_colour



from .context_manager import ContextManager
from .drawable import Drawable
from .thread_generator import ThreadedObject, ThreadedObjectContainer


def check_opengl(checked=False):
    try:
        import OpenGL.GL
    except ImportError as e:
        import os
        import ctypes.util

        uname = os.uname()
        if uname[0] == "Darwin" and uname[2] >= "20." and not checked:
            real_find_library = ctypes.util.find_library

            def find_library(name):
                if name in {"OpenGL", "GLUT"}:  # add more names here if necessary
                    return f"/System/Library/Frameworks/{name}.framework/{name}"
                return real_find_library(name)

            ctypes.util.find_library = find_library
            check_opengl(True)
        else:
            raise e



from amulet_map_editor.api.wx.ui.simple import SimpleChoice, SimpleChoiceAny
import wx
from wx.lib import newevent
import PyMCTranslate
from typing import Tuple, Optional, Type, Any

from amulet.api.data_types import VersionNumberTuple, PlatformType

(
    PlatformChangeEvent,
    EVT_PLATFORM_CHANGE,
) = newevent.NewCommandEvent()  # the platform entry changed
(
    VersionNumberChangeEvent,
    EVT_VERSION_NUMBER_CHANGE,
) = newevent.NewCommandEvent()  # the version number entry changed
(
    FormatChangeEvent,
    EVT_FORMAT_CHANGE,
) = (
    newevent.NewCommandEvent()
)  # the format entry changed (is fired even if the entry isn't visible)
(
    VersionChangeEvent,
    EVT_VERSION_CHANGE,
) = (
    newevent.NewCommandEvent()
)  # one of the above changed. Fired after EVT_FORMAT_CHANGE


class PlatformSelect(wx.Panel):
    def __init__(
        self,
        parent: wx.Window,
        translation_manager: PyMCTranslate.TranslationManager,
        platform: PlatformType = None,
        allow_universal: bool = True,
        allow_vanilla: bool = True,
        allowed_platforms: Tuple[PlatformType, ...] = None,
        **kwargs
    ):
        super().__init__(parent, style=wx.BORDER_SIMPLE)
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._sizer)

        self._translation_manager = translation_manager
        self._allow_universal = allow_universal
        self._allow_vanilla = allow_vanilla
        self._allowed_platforms = allowed_platforms
        self._platform_choice: SimpleChoice = self._add_ui_element(
            "Platform:", SimpleChoice
        )
        self._populate_platform()
        self._set_platform(platform)
        self._platform_choice.Bind(
            wx.EVT_CHOICE,
            lambda evt: wx.PostEvent(
                self, PlatformChangeEvent(self.GetId(), platform=self.platform)
            ),
        )

    def _add_ui_element(
        self, label: str, obj: Type[wx.Control], shown=True, **kwargs
    ) -> Any:
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._sizer.Add(sizer, 0, wx.EXPAND | wx.ALL, 5)
        text = wx.StaticText(self, label=label, style=wx.ALIGN_CENTER)
        sizer.Add(text, 1)
        wx_obj = obj(self, **kwargs)
        sizer.Add(wx_obj, 2)
        if not shown:
            text.Hide()
            wx_obj.Hide()
        return wx_obj

    @property
    def platform(self) -> PlatformType:
        return self._platform_choice.GetCurrentString()

    @platform.setter
    def platform(self, platform: PlatformType):
        self._set_platform(platform)
        wx.PostEvent(self, PlatformChangeEvent(self.GetId(), platform=self.platform))

    def _set_platform(self, platform: PlatformType):
        if platform and platform in self._platform_choice.GetItems():
            self._platform_choice.SetSelection(
                self._platform_choice.GetItems().index(platform)
            )
        else:
            self._platform_choice.SetSelection(0)

    def _populate_platform(self):
        platforms = self._translation_manager.platforms()
        if self._allowed_platforms is not None:
            platforms = [p for p in platforms if p in self._allowed_platforms]
        if not self._allow_universal:
            platforms = [p for p in platforms if p != "universal"]
        if not self._allow_vanilla:
            platforms = [p for p in platforms if p == "universal"]
        self._platform_choice.SetItems(platforms)


class VersionSelect(PlatformSelect):
    def __init__(
        self,
        parent: wx.Window,
        translation_manager: PyMCTranslate.TranslationManager,
        platform: PlatformType = None,
        version_number: VersionNumberTuple = None,
        force_blockstate: bool = None,
        show_force_blockstate: bool = True,
        allow_numerical: bool = True,
        allow_blockstate: bool = True,
        **kwargs
    ):
        super().__init__(parent, translation_manager, platform, **kwargs)
        self._allow_numerical = allow_numerical
        self._allow_blockstate = allow_blockstate

        self.Bind(EVT_PLATFORM_CHANGE, self._on_platform_change)

        self._version_choice: Optional[SimpleChoiceAny] = self._add_ui_element(
            "Version:", SimpleChoiceAny, reverse=True
        )
        self._populate_version()
        self._set_version_number(version_number)
        self._version_choice.Bind(
            wx.EVT_CHOICE,
            lambda evt: wx.PostEvent(
                self,
                VersionNumberChangeEvent(
                    self.GetId(), version_number=self.version_number
                ),
            ),
        )

        self.Bind(EVT_VERSION_NUMBER_CHANGE, self._on_version_number_change)
        self._blockstate_choice: Optional[SimpleChoice] = self._add_ui_element(
            "Format:", SimpleChoice, shown=show_force_blockstate
        )
        self._blockstate_choice.SetItems(["native", "blockstate"])
        self._blockstate_choice.SetSelection(0)
        self._populate_blockstate()
        self._set_force_blockstate(force_blockstate)
        self._blockstate_choice.Bind(
            wx.EVT_CHOICE, lambda evt: self._post_version_change()
        )

    def _post_version_change(self):
        wx.PostEvent(
            self,
            FormatChangeEvent(self.GetId(), force_blockstate=self.force_blockstate),
        ),
        wx.PostEvent(
            self,
            VersionChangeEvent(
                self.GetId(),
                platform=self.platform,
                version_number=self.version_number,
                force_blockstate=self.force_blockstate,
            ),
        )

    @property
    def version_number(self) -> VersionNumberTuple:
        return self._version_choice.GetCurrentObject()

    @version_number.setter
    def version_number(self, version_number: VersionNumberTuple):
        self._set_version_number(version_number)
        wx.PostEvent(
            self,
            VersionNumberChangeEvent(self.GetId(), version_number=self.version_number),
        )

    def _set_version_number(self, version_number: VersionNumberTuple):
        if version_number and version_number in self._version_choice.values:
            self._version_choice.SetSelection(
                self._version_choice.values.index(version_number)
            )
        else:
            self._version_choice.SetSelection(0)

    @property
    def force_blockstate(self) -> bool:
        return self._blockstate_choice.GetCurrentString() == "blockstate"

    @force_blockstate.setter
    def force_blockstate(self, force_blockstate: bool):
        self._set_force_blockstate(force_blockstate)
        self._post_version_change()

    def _set_force_blockstate(self, force_blockstate: bool):
        if force_blockstate is not None:
            self._blockstate_choice.SetSelection(int(force_blockstate))

    @property
    def version(self) -> Tuple[PlatformType, VersionNumberTuple, bool]:
        return self.platform, self.version_number, self.force_blockstate

    @version.setter
    def version(self, version: Tuple[PlatformType, VersionNumberTuple, bool]):
        platform, version_number, force_blockstate = version
        self._set_platform(platform)
        self._set_version_number(version_number)
        self.force_blockstate = force_blockstate

    def _populate_version(self):
        versions = self._translation_manager.version_numbers(self.platform)
        if not self._allow_blockstate:
            versions = [v for v in versions if v < (1, 13, 0)]
        if not self._allow_numerical:
            versions = [v for v in versions if v >= (1, 13, 0)]
        self._version_choice.SetItems(versions)

    def _populate_blockstate(self):
        if self._translation_manager.get_version(
            self.platform, self.version_number
        ).has_abstract_format:
            self._blockstate_choice.Enable()
        else:
            self._blockstate_choice.Disable()

    def _on_platform_change(self, evt):
        self._populate_version()
        self.version_number = None
        evt.Skip()

    def _on_version_number_change(self, evt):
        self._populate_blockstate()
        self.force_blockstate = None
        evt.Skip()


if __name__ == "__main__":

    def main():
        translation_manager = PyMCTranslate.new_translation_manager()
        app = wx.App()
        for cls in (
            PlatformSelect,
            VersionSelect,
            lambda *args: VersionSelect(*args, show_force_blockstate=False),
        ):
            dialog = wx.Dialog(None)
            sizer = wx.BoxSizer()
            dialog.SetSizer(sizer)
            sizer.Add(cls(dialog, translation_manager))
            dialog.Show()
            dialog.Fit()
        app.MainLoop()

    main()



import numpy
from typing import Tuple, List

from amulet.api.chunk import Chunk

from amulet_map_editor.api.opengl.mesh import TriMesh
from amulet_map_editor.api.opengl.resource_pack import (
    OpenGLResourcePackManagerStatic,
    OpenGLResourcePack,
)


try:
    from .chunk_builder_cy import create_lod0_chunk
except:
    raise Exception(
        "Could not import cython chunk mesher. The cython code must be compiled first."
    )


class RenderChunkBuilder(TriMesh):
    """A class to define the logic to generate geometry from a block array"""

    def __init__(self, context_identifier: str, resource_pack: OpenGLResourcePack):
        TriMesh.__init__(self, context_identifier, resource_pack)

    @property
    def chunk(self) -> Chunk:
        raise NotImplementedError

    @property
    def offset(self) -> numpy.ndarray:
        raise NotImplementedError

    @staticmethod
    def _get_block_data(blocks: numpy.ndarray) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """Given a Chunk object will return the chunk arrays needed to generate geometry
        :returns: block array of the chunk, block array one block larger than the chunk, array of unique blocks
        """
        larger_blocks = numpy.zeros(blocks.shape + numpy.array((2, 2, 2)), blocks.dtype)
        larger_blocks[1:-1, 1:-1, 1:-1] = blocks
        unique_blocks = numpy.unique(larger_blocks)
        return larger_blocks, unique_blocks

    def create_geometry(self):
        raise NotImplementedError

    def _set_verts(
        self,
        chunk_verts: List[numpy.ndarray],
        chunk_verts_translucent: List[numpy.ndarray],
    ):
        if chunk_verts:
            self.verts = numpy.concatenate(chunk_verts, None)
            self.verts_translucent = self.verts.size
        else:
            self.verts = self.new_empty_verts()

        if chunk_verts_translucent:
            chunk_verts_translucent.insert(0, self.verts)
            self.verts = numpy.concatenate(chunk_verts_translucent, None)

        self.draw_count = int(self.verts.size // self._vert_len)

    def _create_lod0_multi(
        self, blocks: List[Tuple[numpy.ndarray, int]]
    ) -> Tuple[List[numpy.ndarray], List[numpy.ndarray]]:
        """Create LOD0 geometry data for every sub-chunk in a given chunk.

        :param blocks: A list of tuples containing block arrays extending one block outside the sub-chunk in each direction.
        :return: Opaque block vertices, translucent block vertices.
        """
        return create_lod0_chunk(
            self.resource_pack,
            self.offset,
            blocks,
            self.chunk.block_palette,
            self._vert_len,
        )
